library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
--
-- pragma translate_off
library UNISIM;
use UNISIM.VCOMPONENTS.all;
--pragma translate_on
--
entity FIFO_128 is
     port (
      data_in : in std_logic_vector(7 downto 0);
      clk, reset, write, read : in std_logic;
      data_out : out std_logic_vector(7 downto 0);
      full, data_present, half_full : out std_logic);
end FIFO_128;

architecture arc_FIFO_128 of FIFO_128 is

component SRLC128E_MACRO
  port (
  	D    : in std_logic;
        CE   : in std_logic; 
        clk  : in std_logic;
        A0   : in std_logic;
        A1   : in std_logic;
        A2   : in std_logic;
        A3   : in std_logic;
        A4   : in std_logic;
        A5   : in std_logic;
        A6   : in std_logic;
        Q    : out std_logic;
        Q127  : out std_logic 
       ); 
end component;


signal WRITE_DATA : std_logic_vector(7 downto 0);  
signal CE : std_logic;
signal FIFO_COUNT_internal, ADDRESS_SRL : std_logic_vector(6 downto 0);
signal full_internal, data_present_internal : std_logic;
signal data_out_current, final_data_out : std_logic_vector(7 downto 0);
signal data_out_internal : std_logic_vector(7 downto 0); --:= all zeroes;

begin

WRITE_DATA <= data_in;

-- instantiate SRLC128E_macro
inst_SRLC128E_MACRO : for i in 0 to (7) generate
         U1 :  SRLC128E_MACRO
              port map (
                 D   => WRITE_DATA(i),
                CE   => CE,
                clk  => clk,
                A0   => ADDRESS_SRL(0),
                A1   => ADDRESS_SRL(1),
                A2   => ADDRESS_SRL(2),
                A3   => ADDRESS_SRL(3),
                A4   => ADDRESS_SRL(4),
                A5   => ADDRESS_SRL(5),
                A6   => ADDRESS_SRL(6),
                Q    => data_out_current(i),
                Q127 => final_data_out(i));
       end generate;
      
--address counter
process(clk, reset)
begin
if reset = '1' then
    ADDRESS_SRL <= "0000000";
elsif clk'event and clk = '1' then
    if write = '0' and read = '0' then
        CE <= '0';
        ADDRESS_SRL <= ADDRESS_SRL; --"0000000"; bio bug

    -- read from FIFO
    elsif write = '0' and read = '1' and data_present_internal = '0' then
            CE <= '0';
            ADDRESS_SRL <= ADDRESS_SRL - 1;

    -- write into FIFO
    elsif write = '1' and read = '0' and full_internal = '0' then
           CE <= '1';
        -- increment pointer
           ADDRESS_SRL <= ADDRESS_SRL + 1;

    -- simultaneous read and write into FIFO

    elsif write = '1' and read = '1' and full_internal = '0' and data_present_internal = '0' then
          CE <= '1';
          ADDRESS_SRL <= ADDRESS_SRL;

    else

        CE <= '0';
        ADDRESS_SRL <= ADDRESS_SRL;
        
    end if;
end if;
end process;

-- registered read from FIFO
process(clk, reset)
begin
if reset = '1' then
    data_out_internal <= data_out_internal;
elsif clk'event and clk = '1' then
    if read = '1' and data_present_internal = '0'  then
       data_out_internal <= data_out_current;
    else
       data_out_internal <= data_out_internal;
    end if;
end if;
end process; 

-- count data in the FIFO
process(clk, reset)
begin
if reset = '1' then
    FIFO_COUNT_internal <= "0000000";
elsif clk'event and clk = '1' then
    FIFO_COUNT_internal <= ADDRESS_SRL ;
end if;
end process;

-- Comparator Block
process(clk, reset)
begin
if reset = '1' then
    full_internal <= '1';
elsif clk'event and clk ='1' then
    full_internal <= (ADDRESS_SRL(6) and ADDRESS_SRL(5) and ADDRESS_SRL(4) and ADDRESS_SRL(3) and ADDRESS_SRL(2) and
                       ADDRESS_SRL(1) and not ADDRESS_SRL(0) and write) or
   (ADDRESS_SRL(6) and ADDRESS_SRL(5) and ADDRESS_SRL(4) and ADDRESS_SRL(3) and ADDRESS_SRL(2) and
                       ADDRESS_SRL(1) and ADDRESS_SRL(0) ) ;
end if;
end process;

-- data_present flag is generated when reading from the last location
process(clk, reset)
begin
if reset = '1' then
    data_present_internal <= '1';
elsif clk'event and clk ='1' then
   if (ADDRESS_SRL(6) = '0' and ADDRESS_SRL(5) = '0' and ADDRESS_SRL(4) = '0' and ADDRESS_SRL(3) = '0' and
          ADDRESS_SRL(2) = '0' and ADDRESS_SRL(1) = '0' and ADDRESS_SRL(1) = '0' and read = '1' and write = '0') or
          ( FIFO_COUNT_internal(6) = '0' and FIFO_COUNT_internal(5) = '0' and FIFO_COUNT_internal(4) = '0' and
            FIFO_COUNT_internal(3) = '0' and FIFO_COUNT_internal(2) = '0' and FIFO_COUNT_internal(1) = '0' and
            FIFO_COUNT_internal(0) = '0' and write = '0') then
       data_present_internal <= '1';
   else
       data_present_internal <= '0';
   end if;
             
end if;
end process;

full <= full_internal;
data_present <= data_present_internal;
half_full <= FIFO_COUNT_internal(6);
data_out <= data_out_internal;

end arc_FIFO_128;